#!BPY

# Copyright (c) 2005, Ethan A. Burns <eaburns@unh.edu>
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in
#     the documentation and/or other materials provided with the
#     distribution.
#   * The name of the author may not be used to endorse or promote products
#     derived from this software without specific prior written permission
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
Name: 'Extendable model format (.emf)'
Blender: 232
Group: 'Export'
Tooltip: 'Export selected mesh to a .emf file'
"""

__author__="Ethan Burns"

import Blender
import struct
import cStringIO

hard_factor = 5

# Writes the standard EMF header to the file `f'
def write_emf_header(f):
    data = cStringIO.StringIO()
    data.write(struct.pack("iii", 0x454D4600, 1, 0))
    f.write(data.getvalue())

# Adds the material of this face to the frame
def get_material(mesh, face, frame):
    material = mesh.materials[face.mat]
    sr, sg, sb = material.specCol
    sa = material.specTransp
    r, g, b = material.R, material.G, material.B
    a = material.alpha
    shine = material.hard / hard_factor
    frame.write(struct.pack("ffffffffi", sr, sg, sb, sa, r, g, b, a, shine))

# Adds the indices of a face precided by the number of indices.
def get_indices(face, offset, frame):
    frame.write(struct.pack("i", len(face.v)))
    for i in range(len(face.v)):
        frame.write(struct.pack("i", face.v[i].index+offset))

def get_vert_num():
    retval = 0
    objects = Blender.Object.GetSelected()
    for o in objects:
        mesh = Blender.NMesh.GetRaw(o.data.name)
        if not mesh: continue
        retval += len(mesh.verts)
    return retval

def get_face_num():
    retval = 0
    objects = Blender.Object.GetSelected()
    for o in objects:
        mesh = Blender.NMesh.GetRaw(o.data.name)
        if not mesh: continue
        retval += len(mesh.faces)
    return retval

def get_verts( frame ):
    objects = Blender.Object.GetSelected()
    for o in objects:
        print o
        mesh = Blender.NMesh.GetRaw(o.data.name)
        if not mesh: continue
        for i in range(len(mesh.verts)):
            x, y, z = mesh.verts[i].co
            x += o.LocX
            y += o.LocY
            z += o.LocZ
            print o.LocX
            print o.LocY
            print o.LocZ
            nx, ny, nz = mesh.verts[i].no
            frame.write(struct.pack("ffffff", x, y, z, nx, ny, nz))

def get_faces(frame):
    ind_off = 0
    objects = Blender.Object.GetSelected()
    for o in objects:
        mesh = Blender.NMesh.GetRaw(o.data.name)
        if not mesh: continue
        for i in range(len(mesh.faces)):
            get_material(mesh, mesh.faces[i], frame)
            get_indices(mesh.faces[i], ind_off, frame)
        ind_off += len(mesh.verts)

# Writes an EMF file using the blender API
def write_emf(filename):

    f = open(filename, "wb")

    # Write the EMF header to a file.
    write_emf_header(f)

    # Build the frame
    frame = cStringIO.StringIO()
    num_verts = get_vert_num()
    num_faces = get_face_num()
    frame.write(struct.pack("ii", num_verts, num_faces))

    get_verts(frame)
    get_faces(frame)
    
    # Write a 1 for the number of frames
    tmp = cStringIO.StringIO()
    tmp.write(struct.pack("i", 1))
    
    # Write the frame offset to the file
    # header_len + frame_off_len + num_frames_len = frame_offset
    #     12     +      4        +       4        = 20
    tmp.write(struct.pack("i", 20)) 
    f.write(tmp.getvalue())

    # Write the frame to the file
    f.write(frame.getvalue())

    f.close()
    
def fs_callback(filename):
    if filename.find('.emf', -4) <= 0: filename += '.emf'
    write_emf(filename)

Blender.Window.FileSelector(fs_callback, "Export EMF")
