/*
 * Copyright (c) 2005, Ethan A. Burns <eaburns@unh.edu>
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * The name of the author may not be used to endorse or promote products
 *     derived from this software without specific prior written permission
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <GL/gl.h>
#include <stdint.h>
#include <stdio.h>

#include "emf.h"

#define FRAME_LIST_START 16 /* 16 bytes from start of file to frame list. */

/* Loads the ith frame from the FILE into the buffer. */
static int load_frame( struct emf *buff, unsigned int i, FILE *f )
{
  uint32_t off;
  unsigned int j, k;

  if( i >= buff->nframes ) { 
    fprintf( stderr, "Frame index out of bounds.\n" );
    return -1; 
  }
  
  /* Seek to the frame offset value. */
  if( fseek( f, FRAME_LIST_START+(i*sizeof(uint32_t)), SEEK_SET ) != 0 ) {
    fprintf( stderr, "Error seeking to frame list.\n" );
    return -1;
  }

  if( fread( &off, sizeof(uint32_t), 1, f ) != 1 ) { 
    fprintf( stderr, "Error reading frame offset.\n" );
    return -1; 
  }

  /* Seek to the beginning of the frame. */
  if( fseek( f, off, SEEK_SET ) != 0 ) { 
    fprintf( stderr, "Error seeking to frame offset.\n" );
    return -1; 
  }

  /* Read vertex and face counts. */
  if( fread( &buff->frames[i].nverts, sizeof(uint32_t), 1, f ) != 1 ) {
    fprintf( stderr, "Error reading vertex count.\n" );
    return -1;
  }
  if( fread( &buff->frames[i].nfaces, sizeof(uint32_t), 1, f ) != 1 ) {
    fprintf( stderr, "Error reading frame count.\n" );
    return -1;
  }
  
  /* Allocate vertex list. */
  buff->frames[i].verts = 
    malloc(sizeof(struct emf_vertex)*buff->frames[i].nverts);
  if( buff->frames[i].verts == NULL ) { 
    fprintf( stderr, "Error allocating vertex list: " );
    perror("");
    return -1; 
  }

  /* Allocate face list. */
  buff->frames[i].faces = 
    malloc(sizeof(struct emf_face)*buff->frames[i].nfaces);
  if( buff->frames[i].faces == NULL ) { 
    free( buff->frames[i].verts );
    fprintf( stderr, "Error allocating frame list: " );
    perror("");
    return -1; 
  }

  /* Read the vertices. */
  if( fread( buff->frames[i].verts, sizeof(struct emf_vertex), 
	     buff->frames[i].nverts, f ) != buff->frames[i].nverts ) {
    free( buff->frames[i].verts );
    free( buff->frames[i].faces );
    fprintf( stderr, "Error reading vertices.\n" );
    return -1;
  }
  
  /* Read the faces. */
  for( j = 0; j < buff->frames[i].nfaces; j++ ) {
    if( fread(&buff->frames[i].faces[j].specular, 
	      sizeof(struct emf_color), 1, f) != 1 ) { 
      free( buff->frames[i].verts );
      free( buff->frames[i].faces );
      fprintf( stderr, "Error reading specular color.\n" );
      return -1; 
    }
    if( fread(&buff->frames[i].faces[j].color, 
	      sizeof(struct emf_color), 1, f) != 1 ) { 
      free( buff->frames[i].verts );
      free( buff->frames[i].faces );
      fprintf( stderr, "Error reading face color.\n" );
      return -1; 
    }
    if( fread(&buff->frames[i].faces[j].shininess, 
	      sizeof(uint32_t), 1, f) != 1 ) { 
      free( buff->frames[i].verts );
      free( buff->frames[i].faces );
      fprintf( stderr, "Error reading face shininess.\n" );
      return -1; 
    }
    if( fread(&buff->frames[i].faces[j].n, 
	      sizeof(uint32_t), 1, f ) != 1 ) { 
      free( buff->frames[i].verts );
      free( buff->frames[i].faces );
      fprintf( stderr, "Error reading face index count.\n" );
      return -1; 
    }
    if( fread(buff->frames[i].faces[j].indices, sizeof(uint32_t),
	      buff->frames[i].faces[j].n, f) != buff->frames[i].faces[j].n ) {
      free( buff->frames[i].verts );
      free( buff->frames[i].faces );
      fprintf( stderr, "Error reading face index list.\n" );
      return -1;
    }
  }

  return 0;
}

int load_emf( struct emf *buff, const char *path )
{
  FILE *f;
  uint32_t fid, version;
  unsigned int i;

  f = fopen( path, "rb" );
  if( f == NULL ) {
    fprintf( stderr, "%s: Unable to open file: ", path );
    perror("");
    return -1;
  }

  /* Check the emf file id. */
  fread( &fid, sizeof(uint32_t), 1, f );
  if( fid != 0x454D4600 ) {
    fprintf( stderr, "%s: Error opening file: Invalid file id.\n", path );
    return -1;
  }

  /* Check the version number. */
  fread( &version, sizeof(uint32_t), 1, f );
  if( version > 1 ) {
    fprintf(stderr, "%s: Error opening file: Invalid version number.\n", path);
    return -1;
  }

  /* Read in flags. */
  fread( &buff->flags, sizeof(uint32_t), 1, f );
  
  /* Read in number of frames. */
  fread( &buff->nframes, sizeof(uint32_t), 1, f );
  buff->frames = malloc(sizeof(struct emf_frame)*buff->nframes);
  if( buff->frames == NULL ) {
    fprintf( stderr, "%s: Error opening file: ", path );
    perror("");
    return -1;
  }

  /* Load the frames. */
  for( i = 0; i < buff->nframes; i++ ) {
    if( load_frame( buff, i, f ) != 0 ) {
      free( buff->frames );
      fprintf( stderr, "%s: Error loading frame %d\n", path, i );
      return -1;
    }
  }

  fclose( f );

  /* Set the current frame to frame 0. */
  buff->curframe = buff->frames;

  return 0;
}

void unload_emf( struct emf *buff )
{
  unsigned int i;
  
  for( i = 0; i < buff->nframes; i++ ) {
    free( buff->frames[i].verts );
    free( buff->frames[i].faces );
  }
  
  free( buff->frames );
}

void draw_emf( struct emf *buff )
{
  unsigned int i;

  glEnableClientState( GL_VERTEX_ARRAY );
  glVertexPointer( 3, GL_FLOAT, sizeof(struct emf_vertex), 
		   buff->curframe->verts );

  glEnableClientState( GL_NORMAL_ARRAY );
  glNormalPointer( GL_FLOAT, sizeof(struct emf_vertex), 
		   ((void*)buff->curframe->verts)+(sizeof(float32_t)*3) );

  for( i = 0; i < buff->curframe->nfaces; i++ ) {
    glMaterialfv( GL_FRONT, GL_SPECULAR, 
		  (GLfloat*)&buff->curframe->faces[i].specular );
    glMaterialfv( GL_FRONT, GL_DIFFUSE, 
		  (GLfloat*)&buff->curframe->faces[i].color );
    glMaterialfv( GL_FRONT, GL_AMBIENT, 
		  (GLfloat*)&buff->curframe->faces[i].color );
    glMaterialf( GL_FRONT, GL_SHININESS, buff->curframe->faces[i].shininess );


    glDrawElements( GL_POLYGON, buff->curframe->faces[i].n, GL_UNSIGNED_INT,
		    buff->curframe->faces[i].indices );
  }
}
