#include <floppy.h>
#include <dma.h>
#include <init.h>
#include <bus.h>
#include <cmos.h>
#include <ktypes.h>
#include <kstdio.h>
#include <retval.h>
#include <device.h>
#include <kstdlib.h>
#include <kalloc.h>
#include <queue.h>
#include <irq.h>
#include <timer.h>

static uint32 iot = I386_BUS_IO_T;

/* Floppy controller status register a. */
static uint32 io_stat_a = 0x3F0;
/* Floppy controller status register b. */
static uint32 io_stat_b = 0x3F1;
/* Floppy controller digital output register. */
static uint32 io_dout =   0x3F2; /* Read/Write. */
/* Floppy controller main status register. */
static uint32 io_status = 0x3F4; /* Read Only. */
/* Floppy controller data register. */
static uint32 io_data =   0x3F5; 
/* Floppy controller configuration controll register. (sets transfer rate). */
static uint32 io_cfct =   0x3F7; /* Write Only. */

/* Pointer to the 512 byte floppy drive buffer that is page aligned. */
extern word_t __fd_buffer; /* Defined in arch/i386/kernel.lds. */
static word_t* floppy_buffer = &__fd_buffer;

/* Floppy device structs */
static device_t fd0 = {
  .name = "fd",
  .minor = 0,
  .major = MAJOR_BLOCK_FD
};

static device_t fd1 = {
  .name = "fd",
  .minor = 1,
  .major = MAJOR_BLOCK_FD
};

/* This table should help with figuring out which rate to use:
 * +------------+-----------+----------------+
 * | Drive type | Disk type | Transfer rate  |
 * +------------+-----------+----------------+
 * | 360 KB     | 360 KB    | 250 kbit/s     |
 * +------------+-----------+----------------+
 * | 720 KB     | 720 KB    | 250|300 Kbit/s |
 * +------------+-----------+----------------+
 * | 1.2 MB     | 360 KB    | 300 Kbit/s     |
 * |            +-----------+----------------+
 * |            | 1.2 MB    | 500 Kbit/s     |
 * +------------+-----------+----------------+
 * | 1.44 MB    | 720 MB    | 250 Kbit/s     |
 * |            +-----------+----------------+
 * |            | 1.44 MB   | 500 Kbit/s     |
 * +------------+-----------+----------------+ */

/* Data transfer rates for the cfct register. */
#define CFCT_500 0x0 /* 500 kbit/s */
#define CFCT_300 0x1 /* 300 kbit/s */
#define CFCT_250 0x2 /* 250 kbit/s */
#define CFCT_1M  0x3 /* 1 Mbit/s */

/* Transfer rate table indexed based on FD_T(x)->type */
uint32 fd_trtab[6] = { 0, CFCT_250, CFCT_500, CFCT_300, CFCT_500, CFCT_500 };

/* Floppy controller status bit flags. */
#define FDC_ACTIVE_A 0x1  /* Drive A is active if set. */
#define FDC_ACTIVE_B 0x2  /* Drive B is active if set. */
#define FDC_ACTIVE_C 0x4  /* Drive C is active if set. */
#define FDC_ACTIVE_D 0x8  /* Drive D is active if set. */
#define FDC_BUSY     0x10 /* FDC is busy, don't write a command till clear. */
#define FDC_NDMA     0x20 /* Set if non-dma mode. */
#define FDC_DIO      0x40 /* Direction of IO, set fdc->cpu, clear cpu->fdc. */
#define FDC_MRQ      0x80 /* Set if the data register is not ready for a 
                             read or write. */

/* Floppy digital data register bits. (Low 2 bits are for disk select). */
#define FDC_DIO_RESET   0x04 /* Resets the controller. */
#define FDC_DIO_DMA_IRQ 0x08 /* Disable irq and dma. */
#define FC_DIO_MOTOR_A  0x10 /* When set, drive a's motor is on. */
#define FC_DIO_MOTOR_B  0x10 /* When set, drive b's motor is on. */
#define FC_DIO_MOTOR_C  0x10 /* When set, drive c's motor is on. */
#define FC_DIO_MOTOR_D  0x10 /* When set, drive d's motor is on. */

/* Floppy controller commands from:
 * http://bos.asmhackers.net/docs/floppy/docs/floppy_tutorial.txt */
#define FIX_DRIVE_DATA          0x03
#define CHECK_DRIVE_STATUS      0x04
#define CALIBRATE_DRIVE         0x07
#define CHECK_INTERRUPT_STATUS  0x08
#define FORMAT_TRACK            0x4D
#define READ_SECTOR             0x66
#define READ_DELETE_SECTOR      0xCC
#define READ_SECTOR_ID          0x4A
#define READ_TRACK              0x42
#define SEEK_TRACK              0x0F
#define WRITE_SECTOR            0xC5
#define WRITE_DELETE_SECTOR     0xC9

/* Floppy drive type name strings.  Indexed based on FD_T(x)->type */
char *floppy_type[] = {"No", "360KB 5.25in", "1.2MB 5.25in", 
                       "720KB 3.5in", "1.44MB 3.5in", "2.88MB 3.5in" };
                        
/* Returns the status register's value. */
static __inline__ uint8 fdc_get_status()
{
  uint8 retval;

  bus_space_read_8( iot, io_status, 0, retval );
  return retval;
}

/* Enables the motor on the given drive, 0-3. */
static __inline__ void fdc_enable_motor( uint32 d )
{
  uint8 drive = (0xF&d)<<4;

  bus_space_write_8( iot, io_dout, 0, drive );
}

/* Sets static fd0 and fd1 variables. */
static void floppy_probe()
{
  uint32 cmos_data;

  cmos_data = cmos_get_data( CMOS_FLOPPY_INDEX );  

  FD_T(fd0)->type= (cmos_data>>4)&0xF;
  FD_T(fd1)->type = cmos_data&0xF;
}

/* Interrupt finish function is set to the bottom half of a routine that needs
 * to wait for an interrupt for completion.  If interrupt_finish is NULL, then
 * it is assumed that a routing is not waiting for this interrupt and the 
 * default routine is used. */
static void (*interrupt_finish)() = NULL;

/* Floppy drive interrupt. */
static void fd_interrupt()
{
  if( interrupt_finish != NULL ) { 
    kprintf( "fd: calling interrupt finish.\n" );
    interrupt_finish(); 
    interrupt_finish = NULL;
    return; 
  }

  kprintf( "fd: interrupt default.\n" );
}


/*
 * Reset the floppy controller using a busy-wait.
 */
static int fdc_busy_reset_flag = 0;
static time_event_t busy_reset_timer;
static void fdc_busy_reset_timeout( void* p ) { fdc_busy_reset_flag = 2; }
static void fdc_busy_reset_intr() { fdc_busy_reset_flag = 1; }
static int fdc_busy_reset()
{
  uint32 st, cylinder;
  uint32 i;

  for( i = 0; i < 3; i++ ) {
    fdc_busy_reset_flag = 0;
    time_event_set( &busy_reset_timer, hertz, TIME_EVENT_ONCE, 
                    fdc_busy_reset_timeout, NULL );
    time_event_add( &busy_reset_timer );
  
    interrupt_finish = fdc_busy_reset_intr;
    bus_space_write_8( iot, io_dout, 0, 0x0 );
    bus_space_write_8( iot, io_dout, 0, 0xC );
  
    while( fdc_busy_reset_flag == 0 ) {}
    if( fdc_busy_reset_flag == 2 ) { 
      kprintf( "fd: reset timed out.\n" );
      continue;
    }
    fdc_busy_reset_flag = 1;
  
    bus_space_write_8( iot, io_stat_a, 0, CHECK_INTERRUPT_STATUS );
    bus_space_read_8( iot, io_data, 0, st /*??*/ );
    bus_space_read_8( iot, io_data, 0, cylinder/* cylinder we are at????*/ );

    break;
  }

  if( i == 3 ) { return RET_FAILURE; }
  
  return RET_SUCCESS;
}

/* Called durring device initialization. */
int floppy_init()
{
  uint32 drives = 0;
  ONLY_ONCE();

  fd0.dev_spec = kalloc( sizeof(fd_t) );
  if( fd0.dev_spec == NULL ) {
    kprintf( "fd0: unable to allocate memory for device.\n" );
    return RET_FAILURE;
  }
  fd1.dev_spec = kalloc( sizeof(fd_t) );
  if( fd1.dev_spec == NULL ) {
    kprintf( "fd1: unable to allocate memory for device.\n" );
    return RET_FAILURE;
  }
  floppy_probe();

  /* Print some information about which floppy devices were found. */
  if( FD_T(fd0)->type == 0 && FD_T(fd1)->type == 0 ) { 
    kprintf( "No floppy drives detected.\n" );
    return RET_PROBE_FAILURE;
  }
  if( FD_T(fd0)->type != 0 ) {
    kprintf("fd0: %s floppy drive found.\n", floppy_type[FD_T(fd0)->type] );
    queue_init( fd0.wait_q );
    device_add( &fd0 );
    drives++;
  } else {
    kfree( fd0.dev_spec );
  }

  if( FD_T(fd1)->type != 0 ) {
    kprintf("fd1: %s floppy drive found.\n", floppy_type[FD_T(fd1)->type] );
    queue_init( fd1.wait_q );
    device_add( &fd1 );
    drives++;
  } else {
    kfree( fd1.dev_spec );
  }

  /* If we have at least 1 drive, bind irq6 to the fd_interrupt function. */
  if( drives == 0 ) { return RET_SUCCESS; }

  if( bind_irq( FD_IRQ, fd_interrupt ) == RET_FAILURE ) {
    kprintf( "fd: unable to bind irq 6.\n" );

    kfree( fd0.dev_spec );
    if( drives == 2 )
      kfree( fd1.dev_spec );

    return RET_FAILURE;
  }
  kprintf( "fd: bound irq 6\n" );
  enable_irq( FD_IRQ );

  if( fdc_busy_reset() == RET_FAILURE ) {
    kfree( fd0.dev_spec );
    if( drives == 2 )
      kfree( fd1.dev_spec );

    return RET_FAILURE;
  }

  kprintf( "fd: " );
  bus_space_print( iot, 6, io_stat_a, io_stat_b, io_dout, io_status, 
                   io_data, io_cfct );

  return RET_SUCCESS;
}
init_info( "fd", floppy_init );

/* Initializes the DMA controller for reading from/writing to the floppy disk
 * controller.  Note: the names of the macros are the opposite of the mode for
 * the DMA controller on purpose.  The macro name specifies reading|writing 
 * from/to the fdc, while the DMA mode specifies reading|writing from/to
 * memory. */
#define FD_DMA_SET_READ() \
  dma_program_8( 2, floppy_buffer, 512, DMA_WRITE_TRANSFER|DMA_SINGLE_MODE )
#define FD_DMA_SET_WRITE() \
  dma_program_8( 2, floppy_buffer, 512, DMA_READ_TRANSFER|DMA_SINGLE_MODE )
