#include <kalloc.h>
#include <alloc.h>
#include <kstdio.h>
#include <ktypes.h>
#include <kstdlib.h>
#include <memman.h>
#include <paging.h>
#include <retval.h>

static void* kalloc_base = NULL;

#include <config/kern.h>
#ifdef KALLOC_ROBUST
static uint32 inited = 0;
#endif /* KALLOC_ROBUST */

#ifndef KALLOC_MAX_PAGES
# define KALLOC_MAX_PAGES 16
#endif

static uint32 num_pages;
static word_t pages[KALLOC_MAX_PAGES];

int kalloc_init( void* base, uint32 len )
{
  kalloc_base = base;

#ifdef KALLOC_ROBUST
  inited = 1;
#endif /* KALLOC_ROBUST */

  kprintf( "kalloc: base: 0x%x, len: %d.\n", base, len );
  return init_alloc( base, len );
}

void* kalloc( uint32 size )
{
  word_t* addr;
  uint32 error;

#ifdef KALLOC_ROBUST
  if( inited == 0 ) { 
    kprintf( "WARNING: kalloc used before it is initialized.\n" );
    return NULL; 
  }
#endif /* KALLOC_ROBUST */

  addr = alloc( kalloc_base, size );

  /* If alloc fails, and the memory manager is finished, then we can 
   * grab some more memory for kalloc and try again. */
  if( addr == NULL && memman_inited == 1 && num_pages < KALLOC_MAX_PAGES ) {
    kprintf( "kalloc: allocating new page.\n" );
    pages[num_pages++] = MEMMAN_ALLOC_KERN_FRAME( &error );
    if( error == 1 ) {
      num_pages--;
      kprintf( "kalloc: Error allocating page.\n" );
      return NULL;
    }

    if( init_alloc( (void*)pages[num_pages-1], FRAME_LEN ) == RET_FAILURE ) {
      kprintf( "kalloc: error initializing new alloc page.\n" );
      memman_free_frame( pages[num_pages-1]/FRAME_LEN );
      return NULL;
    }

    alloc_add_free( kalloc_base, (void*)pages[num_pages-1] );
    return alloc( kalloc_base, size );
  }

  return addr;
}

void kfree( void* slot )
{
#ifdef KALLOC_ROBUST
  if( inited == 0 ) { return; }
#endif /* KALLOC_ROBUST */

  if( free( kalloc_base, slot ) == -1 )
    kprintf( "WARNING: Attempting to free unallocated memory in kfree.\n" ); 
}

void* kalloc_get_base() { return kalloc_base; }
