#include <sched.h>
#include <process.h>
#include <ktypes.h>
#include <kstdio.h>
#include <memutil.h>
#include <memspace.h>
#include <memman.h>
#include <kstdlib.h>
#include <timer.h>
#include <kernel.h>
#include <retval.h>
#include <interrupt.h>
#include <cpu.h>
#include <kthread.h>

process_t proctab[SCHED_MAX_PROC];
uint32 curproc = 0;
static time_event_t sched_event;

struct f_queue {
  struct queue q;
  uint32 left;          /* Number left in this round-robbin. */
  uint32 total;         /* Total number in this queue. */
  uint32 quantem;       /* Quantem time for round-robbin. */
  uint32 num;
  struct f_queue *next, *prev;
};

void f_queue_add( struct f_queue *q, process_t *p )
{
  queue_add( &(q->q), p );
  q->total++;
}

process_t *f_queue_pop( struct f_queue *q )
{
  process_t *p = queue_pop( &(q->q) );

  if( p != NULL ) { q->total--; }

  return p;
}

/* Penalty values for the scheduler's argument. */
static int finish_penalty = 1;
static int wait_penalty = -1;
static int no_penalty = 0;
#define PENALTY_NEXT 3

#define SCHED_QUEUES 32
static struct f_queue queues[SCHED_QUEUES];
static struct f_queue *curqueue = queues;

#define SCHED_TIMER_START() \
do { \
  time_event_set( &sched_event, curqueue->quantem, TIME_EVENT_ONCE,  \
                  schedule, &finish_penalty ); \
  time_event_add( &sched_event ); \
} while ( 0 )

/* Returns 1 if the queue has more processes for this round-robbin.
 * Returns 0 if not. 
 * Argument `p' is the last process used. */
static uint32 queue_more( struct f_queue *q, process_t *p )
{
  if( queue_empty( &(q->q) ) == 1 )
    return 0;

  if( q->left == 0 )
    return 0; 

  return 1;
}

/* Adds a process to the current, previous, or next queue depeneding
 * on the current queue it is in, and its penalty value. */
void proc_add_to_queues( process_t *p )
{
  if( curproc == 0 ) { return; }
  if( p->status == PROC_NULL ) { return; }

  if( p->penalty < 0 && curqueue != queues ) {
    p->penalty = 0;
    f_queue_add( curqueue->prev, p );

  } else if( p->penalty >= PENALTY_NEXT && curqueue != queues+SCHED_QUEUES-1 ) {
    p->penalty = 0;
    f_queue_add( curqueue->next, p );

  } else { 
    f_queue_add( curqueue, p );

  }
}

int init_sched()
{
  uint32 i;

  for( i = 1; i < SCHED_MAX_PROC; i++ )
    proctab[i].status = PROC_NULL;

  for( i = 0; i < SCHED_QUEUES; i++ ) {
    queues[i].quantem = i*6+4; /* This gives queue 16 (halfway) a
                                * period of 10 per sec. */
    queues[i].next = queues+i+1;
    queues[i].prev = queues+i-1;
    queues[i].num = i;
    queues[i].total = queues[i].left = 0;
    queue_init( &(queues[i].q) );
  }
  queues[0].prev = queues;
  queues[SCHED_QUEUES-1].next = queues;

  return RET_SUCCESS;
}

void scheduler_start()
{
  SCHED_TIMER_START();
}

void scheduler_stop()
{
  time_event_del( &sched_event );
}

int sched_add( process_t *p )
{
  uint32 i;

  for( i = 1; i < SCHED_MAX_PROC; i++ ) {
    if( proctab[i].status == PROC_NULL ) {
      memcpy( proctab+i, p, sizeof(process_t) );
      proctab[i].status = PROC_RUN;
      proctab[i].pid = i;
      arch_sched_add( proctab+i );
      f_queue_add( queues, proctab+i );
      queues[0].left++;
      return RET_SUCCESS;
    }
  }

  return RET_FAILURE;
}

/* Multi-level feedback queue scheduler implementation, where the argument
 * is a ``penalty'' value used to determine movement between feedback
 * queues. */
void schedule( void* a )
{
  process_t *lastproc = proctab+curproc;
  process_t *p = lastproc;
  int penalty = *(int*)a;
  uint32 waiting_flag = 0;
  uint32 i;
  
  lastproc->penalty += penalty;
  proc_add_to_queues( lastproc );

  /* Start at the first queue, make sure it is finished, before going
   * to the higher level queues. */
  curqueue = queues;

  do {

    for( i = 1; i < SCHED_MAX_PROC && proctab[i].status != PROC_RUN; i++ ) {
      if( proctab[i].status != PROC_NULL ) { waiting_flag = 1; }
    }
    if( proctab[i].status != PROC_RUN ) {
      if( waiting_flag == 1 ) { SCHED_TIMER_START(); enable(); halt(); return; }
      time_event_del( &sched_event );
      kprintf( "\n\nInitial process complete: halting.\n" );
      kend();
    }

    while( queue_more( curqueue, p ) == 1 ) {    
      p = f_queue_pop( curqueue );
      curqueue->left--;
      if( p->status == PROC_RUN ) { goto run; }
      f_queue_add( curqueue, p );
    }

    /* Next queue. */
    curqueue->left = curqueue->total;
    curqueue = curqueue->next;

  } while( /* CONST COND */ 1 );

run:
  curproc = p->pid;
  SCHED_TIMER_START();

  if( lastproc == p ) { return; }

  context_switch( p, lastproc );
}

void process_kill( uint32 pid )
{
  kprintf( "process %d terminating.\n", pid );
  proctab[pid].status = PROC_NULL; 
  if( proctab[pid].kcode == 1 ) {
    kthread_destroy( proctab+pid );
  } else {
    destroy_process( proctab+pid );
  }
  schedule( &no_penalty );
}

void wait( struct queue *wq )
{
  process_t *p = proctab+curproc;

  p->status = PROC_WAIT;
  queue_add( wq, p ); 
  schedule( &wait_penalty );
}

process_t *wake( struct queue *wq )
{
  process_t *p = queue_pop( wq );

  /* Remove all killed processes from the head. */
  while( p != NULL && p->status == PROC_NULL )
    p = queue_pop( wq );

  /* Queue was empty. */
  if( p == NULL ) { return NULL; }

  /* Finally a process that has not been killed.  So lets wake
   * it up. */
  p->status = PROC_RUN; 

  return p;
}
