#if defined(HAVE_CONFIG_H)
#include <config.h>
#endif /* HAVE_CONFIG_H */

#include <sys/ptrace.h>
#include <sys/reg.h>
#include <sys/syscall.h>
#include <sys/types.h>
#include <sys/user.h>
#include <sys/wait.h>

#include <linux/ptrace.h>

#include <err.h>
#include <stdio.h>
#include <stdlib.h>

#include "trace.h"

static int get_syscall_num(const pid_t);
static void *get_path_addr(const pid_t);
static char *get_path(const pid_t, void * const addr);

/*
 * Traces the process with the given pid.  When the given process traps with an
 * open system call entry the open_entry function will be called.  When the
 * given process traps with an open system call exit the open_exit function 
 * will be called.
 *
 * returns negative on error, and non-negative on success.
 */
int trace_syscalls(const pid_t cpid)
{
   int status;			/* The return status of the child. */
   int syscall;			/* The system call number from the child. */
   void *path_addr;		/* Address of the file path arg to open. */
   char *path;			/* Open system call path param. */

   wait(&status);

   while (WIFSTOPPED(status)) {	/* loop while child is stopped. */

      syscall = get_syscall_num(cpid); /* Get the syscall number. */

      if (syscall == SYS_open) {	/* Was it a call to open? */
	 path_addr = get_path_addr(cpid); /* Grab the addr of the path. */
	 path = get_path(cpid, path_addr); /* Grab the path string. */

	 /*
	  * As a demo, just print the open call's path parameter.
	  */
	 if(path != NULL) {
	    fprintf(stderr, "open (%p) %s\n", path_addr, path);
	    free(path);
	 } else {
	    fprintf(stderr, "open (%p)\n", path_addr);
	 }
      }

      /* Let the child process continue until it has a system call. */
      if (ptrace(PTRACE_SYSCALL, cpid, 0, 0) < 0) {
	 warn("%s: ptrace, PTRACE_SYSCALL failed", __func__);
	 return -1;
      }

      wait(&status);
   }

   return 0;
}

/*
 * Sets the current process to allow tracing.
 */
void trace_proc()
{
   ptrace(PTRACE_TRACEME, 0, NULL, NULL);
}

/*
 * Returns the system call number that the process `cpid' trapped on.
 */
static int get_syscall_num(const pid_t cpid)
{
   long orig_rax;

   orig_rax = ptrace(PTRACE_PEEKUSR, cpid, sizeof(long)*ORIG_EAX, NULL);

   return orig_rax;
}

/*
 * Get the address (in the child's memory space) of the path passed to 
 * open.
 */
static void* get_path_addr(const pid_t cpid)
{
   long rdi;

   rdi = ptrace(PTRACE_PEEKUSR, cpid, sizeof(long)*EBX, NULL);

   return (void*)rdi;
}

#define MAX_PATH_LEN 1024

/*
 * Returns the pathname of the file being opened.
 *
 * The return value is alloced and must be freed by the caller.
 */
static char *get_path(const pid_t cpid, void * const paddr)
{
   char *p = paddr;		/* Pointer to the path in cpid's mem. */
   char *path;			/* Return buffer. */
   int i = 0;			/* Index into path buffer. */
   
   if(paddr == NULL) {		/* Make sure we have a good address. */
      warnx("%s: paddr == NULL\n", __func__);
      return NULL;
   }

   /* Allocate the buffer. */
   path = calloc(MAX_PATH_LEN + 1, sizeof(char));

   if(path == NULL)
      err(1, "%s: calloc failed", __func__);

   do {
      path[i] = ptrace(PTRACE_PEEKDATA, cpid, p++, NULL);
   } while (path[i++] != '\0' && i < MAX_PATH_LEN);
   path[i] = '\0';		/* NULL terminate. */

   return path;
}
