/* 
 * uscpfs.c contains the main routine and friends for the uscpfs program.
 * eaburns -- Created Tue Jul 11 18:40:19 2006
 */

#if defined(HAVE_CONFIG_H)
#include <config.h>
#endif /* HAVE_CONFIG_H */

#include <sys/ptrace.h>
#include <sys/types.h>

#include <assert.h>
#include <err.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "debug.h"
#include "trace.h"

/* static function declarations. */
static const char * parseArgs(const int, const char * const[]);
static void printUseage(const char *);
static void execute_cmd(const char *cmd);

/*
 * Entry point.
 */
int main(const int argc, const char * const argv[])
{
   pid_t cpid;
   const char *cmd;

   cmd = parseArgs(argc, argv);

   cpid = fork();

   switch(cpid) {

   case -1:			/* Error. */
      err(1, "%s: fork failed", __func__);

   case 0:			/* Child process */
      /*
       * Execute the child process.
       */
      trace_proc();		/* Enable tracing on this process */
      execute_cmd(cmd);		/* Execute the command. */
      return 1;			/* Returns only on error. */

   default:			/* Parent process */
      /*
       * Begin tracing the child's system calls.
       */
      trace_syscalls(cpid);
   }

   return 0;
}

/*
 * Handles parsing of command line arguments.
 *
 * argc and argv are the parameters to the main function.
 */
static const char * parseArgs(const int argc, const char * const argv[])
{
   if (argc != 2) {
      printUseage(argv[0]);
      return NULL;

   } else {
      return argv[1];
   }
}

/*
 * Prints the useage information.
 *
 * cmd is the command used to run this program (argv[0]).
 */
static void printUseage(const char * cmd)
{
#if defined(HAVE_CONFIG_H)
   printf("%s Useage:\n%s <command>\n", PACKAGE_STRING, cmd);
#else
   printf("Useage:\n%s <command>\n", cmd);
#endif /* HAVE_CONFIG_H */

   exit(0);
}

/*
 * Executes a command where the program is the first space delimited token,
 * and the remaining space delimited tokens are the arguments.
 */
static void execute_cmd(const char *_cmd)
{
   char *argv[4];		/* Argument vector. */
   char *cmd;			/* Copy of the parameter string */

   /*
    * Re-implementation of strdup because aparantly its not yet standard...
    * (copy _cmd into cmd so we have a non-constant).
    */
   cmd = calloc(strlen(_cmd), sizeof(char));

   if(cmd == NULL)
      err(1, "%s: calloc failed", __func__);

   strncpy(cmd, _cmd, strlen(_cmd));

   DPRINTF("cmd=[%s]\n", cmd);

   /*
    * Execute the program with /bin/sh -c <cmd>
    */
   argv[0] = "/bin/sh";
   DPRINTF("argv[%d]=[%s]\n", 0, "/bin/sh");
   argv[1] = "-c";
   DPRINTF("argv[%d]=[%s]\n", 1, "-c");
   argv[2] = cmd;
   DPRINTF("argv[%d]=[%s]\n", 2, cmd);
   argv[3] = NULL;		/* NULL termiante arg vector. */

   /*
    * Execute the command.
    */
   execv(argv[0], argv);
}
