/*
 * Copyright(c) 1997-2001 Id Software, Inc.
 * Copyright(c) 2002 The Quakeforge Project.
 * Copyright(c) 2006 Quetoo.
 * Copyright(c) 2006 MCC.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or(at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
#include <glib.h>
#include <SDL/SDL_ttf.h>

#include "gl_local.h"
#include "console.h"
#include "ref_sdl.h"

/*
 * Texture storing the font-character array. 
 */
static int		fonttex;
static int		char_map = 1;	/* Set to 1 if the charmap is used. */
static unsigned int charmap_width; /* width of fonttex in pixels. */

/*
 * The console font. 
 */
static TTF_Font *con_font = NULL;

/*
 * Console color. 
 */
static SDL_Color con_color = { 255, 255, 255, 200 };	/* white. */

/*
 * NOTE: you need to change CHAR_WIDTH and CHAR_HEIGHT in console.h for
 * new fonts, etc. 
 */
#define CON_FONT_PATH "/usr/local/share/quake2/baseq2/pics/FreeSans.ttf"

extern qboolean scrap_dirty;
void				Scrap_Upload(void);

/*
 * Returns the next closest (larger) value of two. 
 */
static int
nextpoweroftwo(int x)
{
  double logbase2 = log(x) / log(2);
  return (int)((pow(2,ceil(logbase2))) + 0.5);
}


void
Draw_InitLocal(void)
{
	 int				 error;
	 char				c[2];
	 unsigned int	 i,
						  h;
	 SDL_Surface	 *texsurf,
						 *font;
	 SDL_Rect		  destrect;

	 /*
	  * Initialize SDL_ttf 
	  */
	 if (TTF_Init() < 0) {
	fprintf(stderr,
		"Draw_InitLocal: unable to initialize SDL_ttf %s\n",
		SDL_GetError());
	exit(1);
	 }

	 /*
	  * Load the console font. 
	  */
	 fprintf(stderr, "loading console font " CON_FONT_PATH "\n");

	 /*
	  * NOTE: you need to change CHAR_WIDTH and CHAR_HEIGHT in console.h
	  * for new fonts, etc. 
	  */
	 con_font = TTF_OpenFont(CON_FONT_PATH, 16 /* size */ );
	 if (con_font == NULL) {
	fprintf(stderr, "Draw_InitLocal: Unable to load font file "
		CON_FONT_PATH " %s\n", TTF_GetError());
	exit(1);
	 }

	 /*
	  * Create a character map containing a single row of all the characters.
	  * Each element of the character map is CON_CHAR_WIDTHxCON_CHAR_HEIGHT
	  * in size.  Each element contains a single character.
	  */

	 /*
	  * Make sure tex dims are powers of two. 
	  */
	 charmap_width = nextpoweroftwo(CON_CHAR_WIDTH * 128);
	 h = nextpoweroftwo(CON_CHAR_HEIGHT);

	 /*
	  * Create an SDL_Surface that is in a format opengl recognizes to
	  * contain the character map. 
	  */
	 texsurf = SDL_CreateRGBSurface(0, charmap_width, h, 32, 
											  0x00ff0000, 0x0000ff00,
											  0x000000ff, 0xff000000);

	 /*
	  * For each character, blit it to the character map. 
	  */
	 for (i = 1; i < 128; i++) {	/* skip '\0' */
	c[0] = i;
	c[1] = '\0';

	/*
	 * Render the character to a temporary single character surface. 
	 */
#if 0
	font = TTF_RenderText_Solid(con_font, c, con_color);
#else
	SDL_Color		 bg = { 0, 0, 0, 255 };
	font = TTF_RenderText_Shaded(con_font, c, con_color, bg);
#endif
	if (font == NULL) {
		 fprintf(stderr, "Init_DrawLocal: error rendering font %s\n",
			 TTF_GetError());
		 exit(1);
	}

	/*
	 * Find the position for the character in the map. 
	 */
	destrect.x = CON_CHAR_WIDTH * i;
	destrect.y = 0;

	/*
	 * Blit the character to the map. 
	 */
	if (SDL_BlitSurface(font, 0, texsurf, &destrect) == -1) {
		 fprintf(stderr, "Init_DrawLocal: error blitting surface %s\n",
			 SDL_GetError());
		 exit(1);
	}

	SDL_FreeSurface(font);
	 }

	 /*
	  * Create a texture from the character map. 
	  */
	 qglGenTextures(1, &fonttex);
	 qglBindTexture(GL_TEXTURE_2D, fonttex);
	 qglTexImage2D(GL_TEXTURE_2D, 0, 4, charmap_width, h, 0, GL_BGRA, 
						GL_UNSIGNED_BYTE, texsurf->pixels);
	 qglTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	 qglTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

	 /*
	  * Make sure we don't have an error. 
	  */
	 error = qglGetError();
	 if (error != GL_NO_ERROR) {
	fprintf(stderr, "Init_DrawLocal: glError=%d ", error);
	fprintf(stderr, "while loading character map texture: "
		"reverting to on-the-fly font textures... very slow\n");
	qglDeleteTextures(1, &fonttex);
	char_map = 0;
	 }

}

/*
 * Draws a character from the character map.
 */
void
Draw_Char(int x, int y, int num)
{
	 num &= 0xFF;

	 /*
	  * The characters above 128 need to be normalized down to the range
	  * 0-128. 
	  */
	 if (num > 128) {
		num -= 128;
	 }

	 if (num == 32) {		/* space. */
		return;
	 }

	 if (y <= -8)
		return;			// totally off screen

	 if (char_map == 1) {
	float			  smin,
						 smax,
						 tmax;

	/*
	 * Width of the char map texture. 
	 */
	smin = (float) (num * CON_CHAR_WIDTH) / (float)charmap_width;
	smax = (float) ((num + 1) * CON_CHAR_WIDTH) / (float)charmap_width;

	/*
	 * tmax is needed to convert between the actual size of * the
	 * character's glyph and the height of the texture which * is
	 * rounded up to the next power of two. 
	 */
	tmax = (float) CON_CHAR_HEIGHT / nextpoweroftwo(CON_CHAR_HEIGHT);

	qglBindTexture(GL_TEXTURE_2D, fonttex);

	qglBegin(GL_QUADS);
	qglTexCoord2f(smin, 0);
	qglVertex2f(x, y);
	qglTexCoord2f(smax, 0);
	qglVertex2f(x + CON_CHAR_WIDTH, y);
	qglTexCoord2f(smax, tmax);
	qglVertex2f(x + CON_CHAR_WIDTH, y + CON_CHAR_HEIGHT);
	qglTexCoord2f(smin, tmax);
	qglVertex2f(x, y + CON_CHAR_HEIGHT);
	qglEnd();
	qglFlush();

	 } else {

	SDL_Surface	 *font,
						*texsurf;
	int				 texid;
	int				 w,
						 h;
	char				c[2];
	Uint32			 rmask,
						 gmask,
						 bmask,
						 amask;

	c[0] = num;
	c[1] = '\0';

	font = TTF_RenderText_Solid(con_font, c, con_color);

	if (font == NULL) {
		 fprintf(stderr,
			 "Draw_Char_SDL_TTF: unable to render font %s\n",
			 TTF_GetError());
		 exit(1);
	}

	w = nextpoweroftwo(font->w);
	h = nextpoweroftwo(font->h);

	rmask = 0x00ff0000;
	gmask = 0x0000ff00;
	bmask = 0x000000ff;
	amask = 0xff000000;

	texsurf =
		 SDL_CreateRGBSurface(0, w, h, 32, rmask, gmask, bmask, amask);
	SDL_BlitSurface(font, 0, texsurf, 0);

	qglGenTextures(1, &texid);
	qglBindTexture(GL_TEXTURE_2D, texid);
	qglTexImage2D(GL_TEXTURE_2D, 0, 4, w, h, 0, GL_BGRA,
				GL_UNSIGNED_BYTE, texsurf->pixels);

	qglTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	qglTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

	qglBegin(GL_QUADS);
	qglTexCoord2f(0.0f, 0.0f);
	qglVertex2f(x, y);
	qglTexCoord2f(1.0f, 0.0f);
	qglVertex2f(x + w, y);
	qglTexCoord2f(1.0f, 1.0f);
	qglVertex2f(x + w, y + h);
	qglTexCoord2f(0.0f, 1.0f);
	qglVertex2f(x, y + h);
	qglEnd();
	qglFlush();

	SDL_FreeSurface(font);
	SDL_FreeSurface(texsurf);

	qglDeleteTextures(1, &texid);

	 }
}

	/*
	 * Draw_FindPic 
	 */
image_t		  *
Draw_FindPic(char *name)
{
	 image_t		  *gl;
	 char				fullname[MAX_QPATH];

	 if (name[0] != '/' && name[0] != '\\') {
	Com_sprintf(fullname, sizeof(fullname), "pics/%s.pcx", name);
	gl = GL_FindImage(fullname, it_pic);
	 } else
	gl = GL_FindImage(name + 1, it_pic);

	 return gl;
}

	/*
	 * Draw_GetPicSize 
	 */
void
Draw_GetPicSize(int *w, int *h, char *pic)
{
	 image_t		  *gl;

	 gl = Draw_FindPic(pic);
	 if (!gl) {
	*w = *h = -1;
	return;
	 }
	 *w = gl->width;
	 *h = gl->height;
}

	/*
	 * Draw_StretchPic 
	 */
void
Draw_StretchPic(int x, int y, int w, int h, char *pic)
{
	 image_t		  *gl;

	 gl = Draw_FindPic(pic);
	 if (!gl) {
	ri.Con_Printf(PRINT_ALL, "Can't find pic: %s\n", pic);
	return;
	 }

	 if (scrap_dirty)
	Scrap_Upload();

	 if (((gl_config.renderer == GL_RENDERER_MCD) ||
	 (gl_config.renderer & GL_RENDERER_RENDITION)) && !gl->has_alpha) {
	qglDisable(GL_ALPHA_TEST);
	 }

	 GL_Bind(gl->texnum);
	 qglBegin(GL_QUADS);
	 qglTexCoord2f(gl->sl, gl->tl);
	 qglVertex2f(x, y);
	 qglTexCoord2f(gl->sh, gl->tl);
	 qglVertex2f(x + w, y);
	 qglTexCoord2f(gl->sh, gl->th);
	 qglVertex2f(x + w, y + h);
	 qglTexCoord2f(gl->sl, gl->th);
	 qglVertex2f(x, y + h);
	 qglEnd();

	 if (((gl_config.renderer == GL_RENDERER_MCD) ||
	 (gl_config.renderer & GL_RENDERER_RENDITION)) && !gl->has_alpha) {
	qglEnable(GL_ALPHA_TEST);
	 }
}

void
Draw_SolidColor(int x0, int x1, int y0, int y1, const SDL_Color * color)
{
	 if (scrap_dirty) {
	Scrap_Upload();
	 }

	 /*
	  * essentially we're doing a 2D, non textured square. 
	  */
	 qglDisable(GL_TEXTURE_2D);
	 qglDisable(GL_ALPHA_TEST);
	 /*
	  * despite the name, its not unused, its our alpha. 
	  */
	 qglColor4f(color->r, color->g, color->b, color->unused);
	 qglEnable(GL_BLEND);
	 qglDepthMask(GL_FALSE);

	 qglBegin(GL_QUADS);
	 qglVertex2f(x0, y0);
	 qglVertex2f(x0 + x1, y0);
	 qglVertex2f(x0 + x1, y0 + y1);
	 qglVertex2f(x0, y0 + y1);
	 qglEnd();

	 /*
	  * alpha support 
	  */
	 qglDepthMask(GL_TRUE);
	 qglDisable(GL_BLEND);
	 qglColor4f(1, 1, 1, 1);
	 qglEnable(GL_ALPHA_TEST);
	 qglEnable(GL_TEXTURE_2D);
}

void
Draw_Pic(int x, int y, char *pic)
{
	 image_t		  *gl;

	 gl = Draw_FindPic(pic);
	 if (!gl) {
	ri.Con_Printf(PRINT_ALL, "Can't find pic: %s\n", pic);
	return;
	 }
	 if (scrap_dirty) {
	Scrap_Upload();
	 }

	 if (((gl_config.renderer == GL_RENDERER_MCD) ||
	 (gl_config.renderer & GL_RENDERER_RENDITION)) && !gl->has_alpha) {
	qglDisable(GL_ALPHA_TEST);
	 }

	 GL_Bind(gl->texnum);
	 qglBegin(GL_QUADS);
	 qglTexCoord2f(gl->sl, gl->tl);
	 qglVertex2f(x, y);
	 qglTexCoord2f(gl->sh, gl->tl);
	 qglVertex2f(x + gl->width, y);
	 qglTexCoord2f(gl->sh, gl->th);
	 qglVertex2f(x + gl->width, y + gl->height);
	 qglTexCoord2f(gl->sl, gl->th);
	 qglVertex2f(x, y + gl->height);
	 qglEnd();

	 if (((gl_config.renderer == GL_RENDERER_MCD) ||
	 (gl_config.renderer & GL_RENDERER_RENDITION)) && !gl->has_alpha)
	qglEnable(GL_ALPHA_TEST);
}

	/*
	 * Draw_TileClear
	 * 
	 * This repeats a 64*64 tile graphic to fill the screen around a
	 * sized down refresh window. 
	 */
void
Draw_TileClear(int x, int y, int w, int h, char *pic)
{
	 image_t		  *image;

	 image = Draw_FindPic(pic);
	 if (!image) {
	ri.Con_Printf(PRINT_ALL, "Can't find pic: %s\n", pic);
	return;
	 }

	 if (((gl_config.renderer == GL_RENDERER_MCD) ||
	 (gl_config.renderer & GL_RENDERER_RENDITION))
	&& !image->has_alpha)
	qglDisable(GL_ALPHA_TEST);

	 GL_Bind(image->texnum);
	 qglBegin(GL_QUADS);
	 qglTexCoord2f(x / 64.0, y / 64.0);
	 qglVertex2f(x, y);
	 qglTexCoord2f((x + w) / 64.0, y / 64.0);
	 qglVertex2f(x + w, y);
	 qglTexCoord2f((x + w) / 64.0, (y + h) / 64.0);
	 qglVertex2f(x + w, y + h);
	 qglTexCoord2f(x / 64.0, (y + h) / 64.0);
	 qglVertex2f(x, y + h);
	 qglEnd();

	 if (((gl_config.renderer == GL_RENDERER_MCD) ||
	 (gl_config.renderer & GL_RENDERER_RENDITION))
	&& !image->has_alpha) {
	qglEnable(GL_ALPHA_TEST);
	 }
}


	/*
	 * Draw_Fill
	 * Fills a box of pixels with a single color
	 */
void
Draw_Fill(int x, int y, int w, int h, int col)
{
	 union {
	unsigned		  c;
	byte				v[4];
	 } color;

	 if ((unsigned) col > 255)
	ri.Sys_Error(ERR_FATAL, "Draw_Fill: bad color");

	 qglDisable(GL_TEXTURE_2D);

	 color.c = d_8to24table[col];

	 qglColor4ubv(color.v);
	 qglBegin(GL_QUADS);
	 qglVertex2f(x, y);
	 qglVertex2f(x + w, y);
	 qglVertex2f(x + w, y + h);
	 qglVertex2f(x, y + h);
	 qglEnd();
	 qglColor4ubv(color_white);

	 qglEnable(GL_TEXTURE_2D);
}

	/** clears the screen to the defined clear color */
void
Draw_Clear()
{
	 GLbitfield		mask = GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT;

	 if (stencil_buffer == 1)
	mask |= GL_STENCIL_BUFFER_BIT;

	 qglClear(mask);
}


	/*
	 * Draw_FadeScreen
	 * 
	 */
void
Draw_FadeScreen(void)
{
	 qglEnable(GL_BLEND);
	 qglDisable(GL_TEXTURE_2D);
	 qglColor4f(0, 0, 0, 0.8);
	 qglBegin(GL_QUADS);
	 qglVertex2f(0, 0);
	 qglVertex2f(viddef.width, 0);
	 qglVertex2f(viddef.width, viddef.height);
	 qglVertex2f(0, viddef.height);
	 qglEnd();
	 qglColor4ubv(color_white);
	 qglEnable(GL_TEXTURE_2D);
	 qglDisable(GL_BLEND);
}




	/*
	 * Draw_StretchRaw 
	 */
extern unsigned r_rawpalette[256];

void
Draw_StretchRaw(int x, int y, int w, int h, int cols, int rows,
		byte * data)
{
	 unsigned		  image32[256 * 256];
	 unsigned char	image8[256 * 256];
	 int				 i,
						  j,
						  trows;
	 byte			  *source;
	 int				 frac,
						  fracstep;
	 float			  hscale;
	 int				 row;
	 float			  t;

	 GL_Bind(0);

	 if (rows <= 256) {
	hscale = 1;
	trows = rows;
	 } else {
	hscale = rows / 256.0;
	trows = 256;
	 }
	 t = rows * hscale / 256 - 1.0 / 512.0;

	 if (!qglColorTableEXT) {
	unsigned		 *dest;

	for (i = 0; i < trows; i++) {
		 row = (int) (i * hscale);
		 if (row > rows)
		break;
		 source = data + cols * row;
		 dest = &image32[i * 256];
		 fracstep = cols * 0x10000 / 256;
		 frac = fracstep >> 1;
		 for (j = 0; j < 256; j++) {
		dest[j] = r_rawpalette[source[frac >> 16]];
		frac += fracstep;
		 }
	}

	qglTexImage2D(GL_TEXTURE_2D, 0, gl_tex_solid_format, 256, 256, 0,
				GL_RGBA, GL_UNSIGNED_BYTE, image32);
	 } else {
	unsigned char  *dest;

	for (i = 0; i < trows; i++) {
		 row = (int) (i * hscale);
		 if (row > rows)
		break;
		 source = data + cols * row;
		 dest = &image8[i * 256];
		 fracstep = cols * 0x10000 / 256;
		 frac = fracstep >> 1;
		 for (j = 0; j < 256; j++) {
		dest[j] = source[frac >> 16];
		frac += fracstep;
		 }
	}

	qglTexImage2D(GL_TEXTURE_2D,
				0,
				GL_COLOR_INDEX8_EXT,
				256, 256, 0, GL_COLOR_INDEX, GL_UNSIGNED_BYTE,
				image8);
	 }
	 qglTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	 qglTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

	 if ((gl_config.renderer == GL_RENDERER_MCD) ||
	(gl_config.renderer & GL_RENDERER_RENDITION))
	qglDisable(GL_ALPHA_TEST);

	 qglBegin(GL_QUADS);
	 qglTexCoord2f(1.0 / 512.0, 1.0 / 512.0);
	 qglVertex2f(x, y);
	 qglTexCoord2f(511.0 / 512.0, 1.0 / 512.0);
	 qglVertex2f(x + w, y);
	 qglTexCoord2f(511.0 / 512.0, t);
	 qglVertex2f(x + w, y + h);
	 qglTexCoord2f(1.0 / 512.0, t);
	 qglVertex2f(x, y + h);
	 qglEnd();

	 if ((gl_config.renderer == GL_RENDERER_MCD) ||
	(gl_config.renderer & GL_RENDERER_RENDITION))
	qglEnable(GL_ALPHA_TEST);
}
