/*
* Copyright(c) 1997-2001 Id Software, Inc.
* Copyright(c) 2002 The Quakeforge Project.
* Copyright(c) 2006 Quetoo.
* Copyright(c) 2006 MCC.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or(at your option) any later version.
* 
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
* 
* See the GNU General Public License for more details.
* 
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

/*
* Main windowed and fullscreen graphics interface module. This module
* is used for both the software and OpenGL rendering versions of the
* Quake refresh engine.
*/
#include <glib.h>

#ifdef HAVE_CONFIG_H
#   include "config.h"
#endif

#include "client.h"
#include "render.h"
#include "input.h"

// structure containing functions exported from refresh
refexport_t re;

/* Console variables that we need to access from this module */
cvar_t *vid_width;				  /* window size x */
cvar_t *vid_height;				  /* window size y */
cvar_t *vid_fullscreen;			  /* fullscreen on or off */
cvar_t *vid_gamma;				  /* gamma value */

// global video state; used by other modules
viddef_t viddef;

qboolean reflib_active = 0;

// keyboard
void Do_Key_Event(int key, qboolean down);

void (*KBD_Update_fp) (void);
void (*KBD_Init_fp) (Key_Event_fp_t fp);
void (*KBD_Close_fp) (void);

// mouse
in_state_t in_state;

void (*REF_IN_Init_fp) (in_state_t * in_state_p);
void (*REF_IN_Shutdown_fp) (void);
void (*REF_IN_Activate_fp) (qboolean active);
void (*REF_IN_Commands_fp) (void);
void (*REF_IN_Move_fp) (usercmd_t * cmd);
void (*REF_IN_Frame_fp) (void);

void Real_IN_Init(void);

/*
 
REF GLUE
 
*/

#define MAXPRINTMSG	4096
void VID_Printf(int print_level, char *fmt, ...)
{
	va_list argptr;
	char msg[MAXPRINTMSG];

	va_start(argptr, fmt);
	vsnprintf(msg, MAXPRINTMSG, fmt, argptr);
	va_end(argptr);

	if(print_level == PRINT_ALL)
		Com_Printf("%s", msg);
	else
		Com_DPrintf("%s", msg);
}

void __attribute__ ((noreturn)) VID_Error(int err_level, char *fmt, ...)
{
	va_list argptr;
	char msg[MAXPRINTMSG];

	va_start(argptr, fmt);
	vsnprintf(msg, MAXPRINTMSG, fmt, argptr);
	va_end(argptr);

	Com_Error(err_level, "%s", msg);
}


/*
** VID_NewWindow
*/
void VID_NewWindow(int width, int height)
{
	viddef.width = width;
	viddef.height = height;
}

void VID_FreeReflib(void)
{
	if(KBD_Close_fp)
		KBD_Close_fp();
	if(REF_IN_Shutdown_fp)
		REF_IN_Shutdown_fp();

	KBD_Init_fp = NULL;
	KBD_Update_fp = NULL;
	KBD_Close_fp = NULL;
	REF_IN_Init_fp = NULL;
	REF_IN_Shutdown_fp = NULL;
	REF_IN_Activate_fp = NULL;
	REF_IN_Commands_fp = NULL;
	REF_IN_Move_fp = NULL;
	REF_IN_Frame_fp = NULL;

	memset(&re, 0, sizeof(re));
	reflib_active = false;
}

qboolean VID_LoadRefresh()
{
	refimport_t ri;

	/* clean up a previous reflib */
	if(reflib_active) {
		re.Shutdown();
		VID_FreeReflib();
		if(KBD_Close_fp)
			KBD_Close_fp();
		if(REF_IN_Shutdown_fp)
			REF_IN_Shutdown_fp();
		KBD_Close_fp = NULL;
		REF_IN_Shutdown_fp = NULL;
	}

	Com_Printf("loading video driver, ");

	ri.Cmd_AddCommand = Cmd_AddCommand;
	ri.Cmd_RemoveCommand = Cmd_RemoveCommand;
	ri.Cmd_Argc = Cmd_Argc;
	ri.Cmd_Argv = Cmd_Argv;
	ri.Cmd_ExecuteText = Cbuf_ExecuteText;
	ri.Con_Printf = VID_Printf;
	ri.Sys_Error = VID_Error;
	ri.FS_LoadFile = FS_LoadFile;
	ri.FS_FreeFile = FS_FreeFile;
	ri.FS_Gamedir = FS_Gamedir;
	ri.Cvar_Get = Cvar_Get;
	ri.Cvar_Set = Cvar_Set;
	ri.Cvar_SetValue = Cvar_SetValue;
	ri.Vid_NewWindow = VID_NewWindow;

	re = GetRefAPI(ri);

	if(re.api_version != API_VERSION) {
		VID_FreeReflib();
		Com_Error(ERR_FATAL, "video driver has incompatible api_version!");
	}

	/* Init IN(Mouse) */
	in_state.IN_CenterView_fp = IN_CenterView;
	in_state.Key_Event_fp = Do_Key_Event;
	in_state.viewangles = cl.viewangles;
	in_state.in_strafe_state = &in_strafe.state;

	REF_IN_Init_fp = REF_IN_Init;
	REF_IN_Shutdown_fp = REF_IN_Shutdown;
	REF_IN_Activate_fp = REF_IN_Activate;
	REF_IN_Commands_fp = REF_IN_Commands;
	REF_IN_Move_fp = REF_IN_Move;
	REF_IN_Frame_fp = REF_IN_Frame;

	Real_IN_Init();

	if(re.Init(0, 0) == -1) {
		re.Shutdown();
		VID_FreeReflib();
		return false;
	}

	/* Init KBD */
	KBD_Init_fp = KBD_Init;
	KBD_Update_fp = KBD_Update;
	KBD_Close_fp = KBD_Close;

	KBD_Init_fp(Do_Key_Event);

	Key_ClearStates();

	reflib_active = true;
	return true;
}

/*
VID_Restart_f

*/
void VID_Restart_f(void)
{
	vid_width->modified = false;
	vid_height->modified = false;
	vid_fullscreen->modified = false;

	cl.refresh_prepped = false;
	cls.disable_screen = true;

	if(!VID_LoadRefresh()) {	  /* failed */
		Com_Error(ERR_FATAL, "couldn't initialize video\n");
		return;
	}

	cls.disable_screen = false;
}

/*
VID_Init
*/
void VID_Init(void)
{

	Com_Printf("\n------- video initialization -------\n");

	vid_width = Cvar_Get("vid_width", "640", CVAR_ARCHIVE);
	vid_height = Cvar_Get("vid_height", "480", CVAR_ARCHIVE);
	vid_fullscreen = Cvar_Get("vid_fullscreen", "0", CVAR_ARCHIVE);
	vid_gamma = Cvar_Get("vid_gamma", "1", CVAR_ARCHIVE);

	/* Add some console commands that we want to handle */
	Cmd_AddCommand("vid_restart", VID_Restart_f);

	/* Disable the 3Dfx splash screen */
	putenv("FX_GLIDE_NO_SPLASH=0");

	/* Start the graphics mode and load refresher */
	VID_Restart_f();

	Com_Printf("\n");
}

/*
VID_Shutdown
*/
void VID_Shutdown(void)
{
	if(reflib_active) {
		if(KBD_Close_fp)
			KBD_Close_fp();
		if(REF_IN_Shutdown_fp)
			REF_IN_Shutdown_fp();
		KBD_Close_fp = NULL;
		REF_IN_Shutdown_fp = NULL;
		re.Shutdown();
		VID_FreeReflib();
	}
}


/** This is fake, it's actually done by the Refresh load */
void IN_Init(void)
{
}

void Real_IN_Init(void)
{
	if(REF_IN_Init_fp)
		REF_IN_Init_fp(&in_state);
}

void IN_Shutdown(void)
{
	if(REF_IN_Shutdown_fp)
		REF_IN_Shutdown_fp();
}

void IN_Commands(void)
{
	if(REF_IN_Commands_fp)
		REF_IN_Commands_fp();
}

void IN_Move(usercmd_t * cmd)
{
	if(REF_IN_Move_fp)
		REF_IN_Move_fp(cmd);
}

void IN_Frame(void)
{
	if(REF_IN_Activate_fp) {
		if(!cl.refresh_prepped || cls.key_dest == key_console)
			REF_IN_Activate_fp(false);
		else
			REF_IN_Activate_fp(true);
	}

	if(REF_IN_Frame_fp)
		REF_IN_Frame_fp();
}

void IN_Activate(qboolean active)
{
}

void Do_Key_Event(int key, qboolean down)
{
	Key_Event(key, down, Sys_Milliseconds());
}

char *Sys_GetClipboardData(void)
{
	return NULL;
}
